/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Diorite
{

/**
 * InfoBar is a bar designed to show short information.
 */
public class InfoBar: Gtk.Grid
{
	[Description(nick = "Text label", blurb = "Label containing text of the InfoBar.")]
	public Gtk.Label label {get; private set;}
	private Gtk.Button close_button;
	private Gtk.Grid button_box;
	
	/**
	 * Creates new InfoBar.
	 * 
	 * @param icon          icon show on the left
	 * @param text          text of the information
	 * @param use_markup    whether to use Pango markup for the text
	 */
	public InfoBar(string icon, string text, bool use_markup=true)
	{
		Object();
		var icon_image = new Gtk.Image.from_stock(icon, Gtk.IconSize.BUTTON);
		icon_image.halign = Gtk.Align.CENTER;
		icon_image.valign = Gtk.Align.CENTER;
		icon_image.margin = 5;
		icon_image.show();
		add(icon_image);
		
		label = new WrappedLabel(text, use_markup);
		label.xpad = 10;
		label.ypad = 5;
		label.show();
		add(label);
		
		close_button = new Gtk.Button();
		close_button.add(new Gtk.Image.from_stock(Gtk.Stock.CLOSE, Gtk.IconSize.BUTTON));
		close_button.clicked.connect(on_close_button_clicked);
		
		button_box = new Gtk.Grid();
		button_box.expand = false;
		button_box.valign = Gtk.Align.CENTER;
		button_box.add(close_button);
		button_box.show_all();
		add(button_box);
	}
	
	/**
	 * Creates new InfoBar with warning icon.
	 * 
	 * @param text          text of the warning
	 * @param use_markup    whether to use Pango markup for the text
	 */
	public InfoBar.warning(string text, bool use_markup=true)
	{
		this(Gtk.Stock.DIALOG_WARNING, text, use_markup); 
	}
	
	/**
	 * Creates new InfoBar with information icon.
	 * 
	 * @param text          text of the information
	 * @param use_markup    whether to use Pango markup for the text
	 */
	public InfoBar.info(string text, bool use_markup=true)
	{
		this(Gtk.Stock.DIALOG_INFO, text, use_markup);
	}
	
	/**
	 * Creates new InfoBar with error icon.
	 * 
	 * @param text          text of the error
	 * @param use_markup    whether to use Pango markup for the text
	 */
	public InfoBar.error(string text, bool use_markup=true)
	{
		this(Gtk.Stock.DIALOG_ERROR, text, use_markup);
	}
	
	/**
	 * Emitted when the info bar is closed.
	 */
	public virtual signal void closed()
	{
		var parent = get_parent() as Gtk.Container;
		if (parent != null)
			parent.remove(this);
	}
	
	/**
	 * Adds button to info bar
	 * @param button    button to add
	 */
	public void add_button(Gtk.Button button)
	{
		button_box.remove(close_button);
		button_box.add(button);
		button_box.add(close_button);
		button.show();
	}
	
	
	private void on_close_button_clicked()
	{
		closed();
	}
}

} // namespace Diorite
