/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
namespace Diorite.Tools{
	/**
	 * Creates iterator for Flag enums, which iterates from item (1 << 0) to defined
	 * last item. Usage:
	 * 
	 * {{{
	 * [Flags]
	 * public enum Controls{
	 *     PLAY,
	 *     PAUSE,
	 *     PREV,
	 *     NEXT;
	 * 
	 *     public static FlagsIterator<Controls> all(){
	 *         return new FlagsIterator<Controls>(NEXT);
	 *     }
	 * }
	 * 
	 * foreach(Controls c in Controls.all()){
	 *     stdout.printf("Value: %s", c);
	 * }
	 * 
	 * If you wish to skip the first entry meaning NONE, assign a zero to it:
	 * 
	 * [Flags]
	 * public enum Controls{
	 *     NONE = 0,
	 *     PLAY,
	 *     PAUSE,
	 *     PREV,
	 *     NEXT;
	 * }
	 * 
	 * }}}
	 */
	public class FlagsIterator<T>{
        private int index = 0;
        private int step = 0;
        private int last;

        /**
         * Creates new flag enum iterator for enum of type `T` with
         * last entry `last`
         * 
         * @param last the last etry of enum
         */
        public FlagsIterator(T last){
            this.last = (int) last;
        }
		
		/**
		 * Returns self as an iterator. Called by foreach statement.
		 * 
		 * @return self
		 */
		public FlagsIterator<T> iterator(){
			return this;
		}
		
		/**
		 * Availability of the next element.
		 * @return true if there is a next element
		 */
        public bool next() {
            return this.index < this.last;
        }
        
		/**
		 * Returns next element
		 * @return next element
		 */
        public T get() {
//~ 			lib_debug("index before: %d", index);
            this.index = (1 << this.step);
            this.step++;
//~             lib_debug("index after: %d", index);
            return (T) this.index;
        }
    }
    
    /**
	 * Creates iterator for Flag enums, which iterates from item (1 << 0) to defined
	 * last item. Usage:
	 * 
	 * {{{
	 * public enum Fruit{
	 *     BANANA,
	 *     PEAR,
	 *     ORANGE,
	 *     APPLE;
	 * 
	 *     public static EnumIterator<Fruit> all(){
	 *         return new EnumIterator<Fruit>(APPLE);
	 *     }
	 * }
	 * 
	 * foreach(Fruit c in Fruit.all()){
	 *     stdout.printf("Value: %s", c);
	 * }
	 * 
	 * }}}
	 */
	public class EnumIterator<T>{
        private int index = 0;
        private int last;

        /**
         * Creates new flag enum iterator for enum of type T with
         * last entry last
         * 
         * @param last the last etry of enum
         */
        public EnumIterator(T last){
            this.last = (int) last;
        }
		
		/**
		 * Returns self as an iterator. Called by foreach statement.
		 * 
		 * @return self
		 */
		public EnumIterator<T> iterator(){
			return this;
		}
		
		/**
		 * Availability of the next element.
		 * @return true if there is a next element
		 */
        public bool next() {
            return this.index <= this.last;
        }
        
		/**
		 * Returns next element
		 * @return next element
		 */
        public T get() {
            return (T) (this.index++);
        }
    }
}
