/*
 * Copyright 2012-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
using Diorite.System;

/**
 * Module providing lyrics fetching
 */
namespace Nuvola.Extensions.Lyrics{

private const int DEFAULT_WIDTH = 300;

const string UI_SPEC = """
<ui>
	<menubar action="menubar">
		<menu action="view">
			<placeholder name="view-bottom">
				<menuitem action="sidebar-toggle" />
			</placeholder>
		</menu>
	</menubar>
</ui>
""";

namespace Actions
{
	public const string TOGGLE_SIDEBAR = "sidebar-toggle";
}

/**
 * Returns plugin information
 * 
 * @return plugin information
 */
public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing lyrics fetching
		_("Lyrics fetching"),
		Config.VERSION,
		/// Description of a plugin providing lyrics fetching
		_("<p>This plugin provides fetching of lyrics from <a href=\"http://www.azlyrics.com/\">AZ Lyrics</a>.</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

/**
 * Escapes name for safe usage in file names
 * 
 * @param name	string to escape
 * @return		escaped name
 */
public string escape_name(string name){
	return GLib.Uri.escape_string(name, " ").replace("%", ",");
}

/**
 * Errors in lyrics fetching process
 */
public errordomain Error{
	/**
	 * Artist and song pain leads to invalid id
	 */
	WRONG_ID,
	/**
	 * Lyrics has not been found
	 */
	NOT_FOUND
}

/**
 * Interface for lyrics fetchers
 */
public interface Fetcher: Object{
	/**
	 * Fetch lyrics
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 * @throws Error     on failure
	 */ 
	public abstract async string fetch_lyrics(string artist, string song) throws Error;
}


/**
 * Lyrics fetching plugin
 */
public class Extension: Nuvola.Extension{
	[Description(nick = "View visibility", blurb = "Whether the lyrics view is visible")]
	private bool view_visible{
		get{
			return ui_state.get_bool(EXTENSIONS_KEY + "." + id + ".visible", true);
		}
		set{
			ui_state.set_bool(EXTENSIONS_KEY + "." + id + ".visible", value);
		}
	}
	
	[Description(nick = "View width", blurb = "With of the lyrics view.")]
	private int view_width{
		get{
			return ui_state.get_int(EXTENSIONS_KEY + "." + id + ".view_width", DEFAULT_WIDTH);
		}
		set{
			ui_state.set_int(EXTENSIONS_KEY + "." + id + ".view_width", value);
		}
	}
	
	private weak Diorite.MultiTypeMap ui_state;
	private weak Nuvola.Player player;
	private weak Diorite.Actions actions;
	public Diorite.ToggleAction toggle_action;
	private CachedLyrics cache;
	private SList<Fetcher> fetchers;
	private View? view = null;
	private uint ui_merge_id = 0;
	private weak MainWindow? main_window;
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws Nuvola.ExtensionError{
		// Get dependencies, throws ExtensionError
		player = objects.get<Player>("player");
		main_window = objects.get<MainWindow>("main_window");
		var storage = objects.get<Diorite.Storage>("storage");
		actions = objects.get<Diorite.Actions>("actions");
		ui_state = objects.get<Diorite.MultiTypeMap>("ui_state");
		append_ui_actions();
		
		cache = new CachedLyrics(storage);
		fetchers = new SList<Fetcher>();
		fetchers.append(cache);
		fetchers.append(new AZLyrics());
		
		view = new View(this);
		if(view_visible) view.show();
		
		main_window.set_sidebar(view);
		Gtk.Allocation allocation;
		main_window.paned.get_allocation (out allocation);
		main_window.paned.position = allocation.width - view_width;
		main_window.paned.position_set = true;
		main_window.paned.notify["position"].connect(
			on_view_position_changed);
		
		player.song_changed.connect(on_song_changed);
		view.refresh_lyrics.connect(queue_fetch_lyrics);			
		view.save_lyrics.connect(cache.store);
		
		if(player.song != null && player.artist != null){
			queue_fetch_lyrics(player.artist, player.song);
		}
		
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void add_actions_ui(Gtk.UIManager manager)
	{
		try
		{
			ui_merge_id = manager.add_ui_from_string(UI_SPEC, UI_SPEC.length);
		}
		catch (GLib.Error e)
		{
			warning("[%s] Unable to add ui: %s", id, e.message);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void remove_actions_ui(Gtk.UIManager manager){
		if(ui_merge_id != 0){
			manager.remove_ui(ui_merge_id);
			ui_merge_id = 0;
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		debug("Extensions[%s] unload", id);
		player.song_changed.disconnect(on_song_changed);
		main_window.paned.notify["position"].disconnect(
			on_view_position_changed);
		main_window.paned.remove(view);
		actions.remove_action(Actions.TOGGLE_SIDEBAR);
	}
	
	/**
	 * Callback for song change. Queues Lyrics fetching.
	 */
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		if(song == null || artist == null){
			return;
		}
		queue_fetch_lyrics(artist, song);
	}
	
	private void on_view_position_changed(GLib.Object o, ParamSpec p){
		if(!view.visible) return;
		Gtk.Allocation allocation;
		view.get_allocation (out allocation);
		view_width = allocation.width;
	}
	
	private void queue_fetch_lyrics(string artist, string song){
		lyrics_loading(artist, song);
		fetch_lyrics.begin(artist, song);
	}
		
	private async void fetch_lyrics(string artist, string song){
		foreach(var fetcher in fetchers){
//~ 				debug("Fetcher, %s", fetcher.get_type().name());
			try{
				var lyrics = yield fetcher.fetch_lyrics(artist, song);
				if(fetcher != cache){
					cache.store(artist, song, lyrics);
				}
				Idle.add(() =>{
					lyrics_available(artist, song, lyrics);
					return false;
				});
				return;
			}
			catch(Error e){
				debug("Fetch error: %s", e.message);
			}
		}
		
		lyrics_not_found(artist, song);
	}
	
	private void append_ui_actions(){
		return_if_fail(actions != null);
		/// Label for toggle sidebar action
		toggle_action = new Diorite.ToggleAction(Actions.TOGGLE_SIDEBAR, null, _("Show sidebar"), null, null, view_visible);
		toggle_action.toggled.connect(() => {
			view.visible = view_visible = toggle_action.active;
		});
		actions.add_action(toggle_action);
	}
	
	/**
	 * Emitted when lyrics is available.
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 * @param lyrics    lyrics text
	 */
	public signal void lyrics_available(string artist, string song, string lyrics);
	
	/**
	 * Emitted when lyrics has not been found.
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 */
	public signal void lyrics_not_found(string artist, string song);
	
	/**
	 * Emitted when lyrics fetching was enqueued. 
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 */
	public signal void lyrics_loading(string artist, string song);
}

/**
 * File-based cache for lyrics
 */
public class CachedLyrics : Object, Fetcher{
	private const string FILE_FORMAT = "lyrics/%s/%s.txt";
	private weak Diorite.Storage storage;
	
	/**
	 * Creates new lyrics cache
	 * 
	 * @param storage    where to store lyrics
	 */
	public CachedLyrics(Diorite.Storage storage){
		this.storage = storage;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public async string fetch_lyrics(string artist, string song) throws Error{
		var fixed_artist = escape_name(artist.down());
		var fixed_song = escape_name(song.down());
		
		if(fixed_artist == "" || fixed_song == ""){
			throw new Error.NOT_FOUND(@"Song $song was not found in cache");
		}
		 
		var cached = storage.get_cache_path(FILE_FORMAT.printf(fixed_artist, fixed_song));
		if(cached.query_exists(null)){
			try{
				var lyrics = Diorite.System.read_file(cached);
				if(lyrics != null && lyrics != ""){
					return lyrics;
				}
			}
			catch(GLib.Error e){
				warning("Unable to load cached lyrics: %s", e.message);
				throw new Error.NOT_FOUND(@"Unable to load song $song from cache");
			}
		}
		throw new Error.NOT_FOUND(@"Song $song was not found in cache");
	}
	
	/**
	 * Store lyrics to cache.
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 * @param lyrics    lyrics text
	 */
	public void store(string artist, string song, string lyrics){
		var fixed_artist = escape_name(artist.down());
		var fixed_song = escape_name(song.down());
		
		if(fixed_artist == "" || fixed_song == ""){
			return;
		}
		 
		var cached = storage.get_cache_path(FILE_FORMAT.printf(fixed_artist, fixed_song));
		
		try{
			cached.get_parent().make_directory_with_parents();
		}
		catch(GLib.Error e){}
		try{
			overwrite_file(cached, lyrics);
		}
		catch(GLib.Error e){
			warning("Unable to store lyrics: %s", e.message);
		}
	}
}

/**
 * Lyric fetcher for [[http://www.azlyrics.com/|AZ Lyrics]].
 */
public class AZLyrics : Object, Fetcher{
	/**
	* URL format of AZ Lyrics
	*/
	private const string SONG_PAGE = "http://www.azlyrics.com/lyrics/%s/%s.html";
	private Regex html_tags;
	
	construct{
		
		try{
			html_tags = new Regex("</?\\w+?( /)?>", RegexCompileFlags.CASELESS);
		}
		catch(RegexError e){
			error("RegexError: %s", e.message);
		}
	}
	/**
	 * {@inheritDoc}
	 */
	public async string fetch_lyrics(string artist, string song) throws Error{
		Soup.Message message;
		var url = SONG_PAGE.printf(transform_name(artist), transform_name(song));
		debug("AZLyrics: %s", url);
		message = new Soup.Message("GET", url);
		var session = WebKit.get_default_session();
		
		SourceFunc callback = fetch_lyrics.callback;
		session.queue_message(message, () => {
			Idle.add((owned) callback);
		});
		yield;
		
		var response = (string) message.response_body.flatten().data;
		if(message.status_code != 200 || response == ""){
			throw new Error.NOT_FOUND(@"Song $song was not found on AZ Lyrics");
		}
		
		response = parse_response(response);
		if(response == ""){
			throw new Error.NOT_FOUND(@"Song $song was not found on AZ Lyrics");
		}
		
		return response;
	}
	
	private string parse_response(string response){
		const string START = "<!-- start of lyrics -->";
		const string END = "<!-- end of lyrics -->";
		var start_pos = response.index_of(START);
		if(start_pos >= 0){
			var end_pos = response.index_of(END);
			if(end_pos >= 0 && end_pos > start_pos){
				var lyrics = response.slice(start_pos + START.length, end_pos);
				try{
					lyrics = html_tags.replace_literal(lyrics, lyrics.length, 0, "");
				}
				catch(RegexError e){
					warning("RegexError: %s", e.message);
				}
				return lyrics.strip() + "\n";
			}
		}
		return "";
	}
	
	/**
	 * Transforms name to match AZLyrics identifiers
	 * 
	 * @param name		original name
	 * @return			transformed name
	 */
	public static string transform_name(string name){
		var normalized = name.normalize();
		var buffer = new StringBuilder("");
		unichar c;
		int i = 0;
		while(normalized.get_next_char(ref i, out c)){
			c = c.tolower();
			if(('a' <= c && c <= 'z') || ('0' <= c && c <= '9')){
				buffer.append_unichar(c);
			}
		}
		
		return buffer.str;
	}
}

/**
 * Lyrics Extension View is GUI showing lyrics.
 */
public class View: Gtk.Grid{
	private weak Extension plugin;
	private Gtk.Label header;
	private Gtk.Label status;
	private Gtk.TextView view;
	private string? artist = null;
	private string? song = null;
	private Gtk.Grid bar;
	
	/**
	 * Creates new View
	 */
	public View(Extension plugin){
		row_spacing = 5;
		column_homogeneous = false;
		this.plugin = plugin;
		var text = "<b>%s</b>".printf(Markup.escape_text(_("Song lyrics")));
		header = new Gtk.Label(null);
		header.set_markup((owned) text);
		header.hexpand = true;
		header.halign = Gtk.Align.FILL;
		attach(header, 0, 0, 1, 1);
		
		var close_button = new Gtk.Button();
		close_button.add(new Gtk.Image.from_stock(Gtk.Stock.CLOSE, Gtk.IconSize.BUTTON));
		close_button.clicked.connect(() => { plugin.toggle_action.activate(); });
		close_button.expand = false;
		close_button.relief = Gtk.ReliefStyle.NONE;
		close_button.halign = Gtk.Align.END;
		close_button.show_all();
		attach(close_button, 1, 0, 1, 1);
		
		status = new Gtk.Label(null);
		status.no_show_all = true;
		attach(status, 0, 1, 2, 1);
		
		view = new Gtk.TextView();
		view.left_margin = 5;
		view.right_margin = 5;
		view.pixels_above_lines = 1;
		view.pixels_below_lines = 1;
		view.wrap_mode = Gtk.WrapMode.WORD;
		view.buffer.modified_changed.connect(on_buffer_changed);
		
		var scroll = new Gtk.ScrolledWindow(null, null);
		scroll.vexpand = true;
		scroll.add(view);
		scroll.expand = true;
		attach(scroll, 0, 2, 2, 1);
		
		var refresh_button = new Gtk.Button.from_stock(Gtk.Stock.REFRESH);
		refresh_button.hexpand = true;
		refresh_button.clicked.connect(() => {
			if(artist == null || song == null) return;
			refresh_lyrics(artist, song);
		});
		refresh_button.show();
		
		var save_button = new Gtk.Button.from_stock(Gtk.Stock.SAVE);
		save_button.hexpand = true;
		save_button.clicked.connect(() => {
			if(artist == null || song == null) return;
			save_lyrics(artist, song, view.buffer.text);
		});
		save_button.show();
		
		bar = new Gtk.Grid();
		bar.orientation = Gtk.Orientation.HORIZONTAL;
		bar.no_show_all = true;
		bar.add(refresh_button);
		bar.add(save_button);
		attach(bar, 0, 3, 2, 1);
		
		foreach (var child in get_children())
			child.hexpand = true;
		
		plugin.lyrics_available.connect(on_lyrics_available);
		plugin.lyrics_not_found.connect(on_lyrics_not_found);
		plugin.lyrics_loading.connect(on_lyrics_loading);
		header.show();
		scroll.show_all();
	}
	
	private void on_lyrics_available(string artist, string song, string lyrics){
//~ 			debug("Lyrics: %s/%s: %s", artist, song, lyrics);
		this.artist = artist;
		this.song = song;
		status.set_text(song);
		view.buffer.set_text(lyrics);
		view.buffer.set_modified(false);
		view.editable = true;
	}
	
	private void on_lyrics_not_found(string artist, string song){
		this.artist = artist;
		this.song = song;
		status.set_text(_("No lyrics has been found."));
		status.show();
		view.buffer.set_text("");
		view.buffer.set_modified(false);
		view.editable = true;
	}
	
	private void on_lyrics_loading(string artist, string song){
		this.artist = artist;
		this.song = song;
		status.set_text(_("Fetching lyrics ..."));
		status.show();
		view.buffer.set_text("");
		view.buffer.set_modified(false);
		view.editable = false;
	}
	
	private void on_buffer_changed(Gtk.TextBuffer buffer){
		bar.visible = buffer.get_modified();
	}
	
	/**
	 * Emitted when user requests to save lyrics
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 * @param lyrics    lyrics text
	 */ 
	public virtual signal void save_lyrics(string artist, string song, string lyrics){
		view.buffer.set_modified(false);
	}
	
	/**
	 * Emitted when user requests to refresh lyrics
	 * 
	 * @param artist    artist name
	 * @param song      song name
	 */ 
	public signal void refresh_lyrics(string artist, string song);
	
}

} // namespace Nuvola.Extensions.Lyrics
