/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola
{

public class DownloadsDialog: Gtk.Window
{
	private Gtk.Grid grid;
	private Gtk.Button clear_button;
	private SList<Gtk.Widget> finished = null;
	
	/**
	 * Creates new Downloads dialog
	 */
	public DownloadsDialog(Diorite.Application app)
	{
		Object(window_position: Gtk.WindowPosition.CENTER);
		set_default_size(400, 300);
		/// Title of downloads window, %s will be replaced by application name
		title = _("Downloads - %s").printf(app.display_name);
		try
		{
			icon = Gtk.IconTheme.get_default().load_icon(app.icon, 48, 0);
		}
		catch(Error e)
		{
			warning("Unable to load application icon.");
		}
		var main_box = new Gtk.Grid();
		main_box.orientation = Gtk.Orientation.VERTICAL;
		var view = new Gtk.ScrolledWindow(null, null);
		grid = new Gtk.Grid();
		grid.orientation = Gtk.Orientation.VERTICAL;
		grid.show();
		view.add_with_viewport(grid);
		view.vexpand = true;
		view.show();
		main_box.add(view);
		
		var button_box = new Gtk.ButtonBox(Gtk.Orientation.HORIZONTAL);
		button_box.margin = 5;
		button_box.layout_style = Gtk.ButtonBoxStyle.EDGE;
		button_box.hexpand = true;
		button_box.halign = Gtk.Align.FILL;
		clear_button = new Gtk.Button.from_stock(Gtk.Stock.CLEAR);
		clear_button.clicked.connect(on_clear_button);
		clear_button.show();
		button_box.add(clear_button);
		var close_button = new Gtk.Button.from_stock(Gtk.Stock.CLOSE);
		clear_button.sensitive = false;
		close_button.clicked.connect(on_close_button);
		close_button.show();
		button_box.add(close_button);
		button_box.show_all();
		main_box.add(button_box);
		main_box.show();
		add(main_box);
	}
	
	public void add_download(WebKit.Download request, File destination)
	{
		var download = new Download(request, destination);
		download.finished.connect(on_download_finished);
		grid.insert_row(0);
		grid.attach(download, 0, 0, 1, 1);
		download.show();
	}
	
	public override bool delete_event(Gdk.EventAny event)
	{
		hide();
		return true;
	}
	
	private void on_download_finished(Download download)
	{
		finished.prepend(download);
		download.finished.disconnect(on_download_finished);
		clear_button.sensitive = true;
	}
	
	private void on_clear_button()
	{
		foreach (var widget in finished)
			grid.remove(widget);
		finished = null;
		clear_button.sensitive = false;
	}
	
	private void on_close_button()
	{
		hide();
	}
	
	private class Download: Gtk.Frame
	{
		private WebKit.Download request;
		private File destination;
		private Gtk.ProgressBar? progress_bar;
		private Gtk.Label? bare_label;
		private Gtk.Grid? box;
		
		public Download(WebKit.Download request, File destination)
		{
			Object(shadow_type: Gtk.ShadowType.OUT);
			hexpand = true;
			this.request = request;
			this.destination = destination;
			progress_bar = new Gtk.ProgressBar();
			progress_bar.text = destination.get_basename();
			progress_bar.show_text = true;
			progress_bar.margin_right = 10;
			progress_bar.hexpand = true;
			bare_label = null;
			box = new Gtk.Grid();
			box.margin = 5;
			box.orientation = Gtk.Orientation.HORIZONTAL;
			box.add(progress_bar);
			var cancel = new Gtk.Button.with_label(_("Cancel download"));
			cancel.clicked.connect(on_cancel_download);
			box.add(cancel);
			add(box);
			box.show_all();
			request.notify["status"].connect_after(on_status_changed);
		}
		
		public signal void finished();
		
		private void update()
		{
			if (progress_bar != null)
				progress_bar.fraction = request.progress;
		}
		
		private void on_status_changed(GLib.Object o, ParamSpec p)
		{
			debug("Download status: %s", request.status.to_string());
			switch (request.status)
			{
			case WebKit.DownloadStatus.STARTED:
				request.notify["current-size"].connect_after(on_progress_changed);
				update();
				break;
			case WebKit.DownloadStatus.FINISHED:
				request.notify["current-size"].disconnect(on_progress_changed);
				set_bare_label();
				var open_file = new Gtk.Button.with_label(_("Open file"));
				open_file.clicked.connect(on_open_file);
				box.add(open_file);
				var open_dir = new Gtk.Button.with_label(_("Open directory"));
				open_dir.clicked.connect(on_open_dir);
				box.add(open_dir);
				box.show_all();
				finished();
				break;
			case WebKit.DownloadStatus.ERROR:
				request.notify["current-size"].disconnect(on_progress_changed);
				set_bare_label(_("Download failed"));
				finished();
				break;
			case WebKit.DownloadStatus.CANCELLED:
				request.notify["current-size"].disconnect(on_progress_changed);
				set_bare_label(_("Download canceled"));
				finished();
				break;
			default:
				warning("Unhandled status: %s", request.status.to_string());
				break;
			}
		}
		
		private void set_bare_label(string? other_label=null)
		{
			bare_label = new Gtk.Label(progress_bar.text);
			bare_label.margin_right = 10;
			bare_label.hexpand = true;
			bare_label.halign = Gtk.Align.START;
			progress_bar = null;
			foreach (var child in box.get_children())
				box.remove(child);
			box.add(bare_label);
			if (other_label != null)
				box.add(new Gtk.Label(other_label));
			box.show_all();
		}
		
		private void on_progress_changed(GLib.Object o, ParamSpec p)
		{
			update();
		}
		
		private void on_cancel_download()
		{
			request.cancel();
		}
		
		private void on_open_file()
		{
			Diorite.System.open_uri_with_fallback(destination.get_path());
		}
		
		private void on_open_dir()
		{
			Diorite.System.open_uri_with_fallback(destination.get_parent().get_path());
		}
	}
}

} // namespace Nuvola
