/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Nuvola{

/**
 * ExtensionManagerView provided graphicas interface for extensions management.
 */
public class ExtensionsManagerView: Gtk.Grid{
	private Gtk.TreeView selector;
	private Gtk.ListStore model;
	private weak ExtensionsManager extensions;
	private Extension? extension = null;
	private weak Diorite.MultiTypeMap config;
	private Gtk.TextView description;
	private Gtk.Button settings_button;
	
	/**
	 * Creates new ExtensionsManagerView
	 * 
	 * @param extensions	extensions manager
	 * @param config		configuration
	 */
	public ExtensionsManagerView(ExtensionsManager extensions, Diorite.MultiTypeMap config){
		this.extensions = extensions;
		this.config = config;
		
		// id, enabled, name
		model = new Gtk.ListStore(EMI.COUNT, typeof(string), typeof(bool), typeof(string));
		var keys = extensions.available_extensions.get_keys();
		keys.sort(strcmp);
		Gtk.TreeIter iter;
		foreach(var key in keys){
			var extension = extensions.available_extensions.lookup(key);
			var enabled = extensions.get(key) != null;
			model.append(out iter);
			model.set(iter, EMI.ID, key, EMI.ACTIVE, enabled, EMI.NAME, extension.name);
		}
		
		var win = new Gtk.ScrolledWindow(null, null);
		selector = new Gtk.TreeView.with_model(model);
		selector.headers_clickable = true;
		win.add(selector);
		win.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		win.shadow_type = Gtk.ShadowType.IN;
		win.vexpand = true;
		border_width = 3;
		win.border_width = 3;
		var left_box = new Gtk.Grid();
		left_box.orientation = Gtk.Orientation.VERTICAL;
		left_box.vexpand = true;
		left_box.add(win);
		add(left_box);
		
		var enabled_renderer = new Gtk.CellRendererToggle();
		enabled_renderer.toggled.connect(on_extension_toggled);
		var enabled_column = new Gtk.TreeViewColumn.with_attributes(
			/// header in extensions manager
				_("Enabled"), enabled_renderer, "active", EMI.ACTIVE);
		var name_renderer = new Gtk.CellRendererText();
		var name_column = new Gtk.TreeViewColumn.with_attributes(
			/// header in extensions manager
				_("Extension name"), name_renderer, "text", EMI.NAME);
		name_column.sort_column_id = EMI.NAME;
		selector.append_column(enabled_column);
		selector.append_column(name_column);
		
		
		var right_box = new Gtk.Grid();
		right_box.orientation = Gtk.Orientation.VERTICAL;
		right_box.vexpand = true;
		add(right_box);
		description = new Diorite.SimpleDocView();
		description.border_width = 10;
		description.editable = false;
		description.wrap_mode = Gtk.WrapMode.WORD_CHAR;
		
		var scroll_box = new Gtk.ScrolledWindow(null, null);
		scroll_box.vexpand = true;
		scroll_box.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		win.shadow_type = Gtk.ShadowType.IN;
		right_box.add(scroll_box);
		
		scroll_box.add(description);
		scroll_box.set_size_request(250, -1);
		
		settings_button = new Gtk.Button.from_stock(Gtk.Stock.PREFERENCES);
		right_box.add(settings_button);
		settings_button.sensitive = false;
		settings_button.clicked.connect(on_settings_button_clicked);
		
		var selection = selector.get_selection();
		selection.changed.connect(on_selection_changed);
		selection.set_mode(Gtk.SelectionMode.SINGLE);
		if(model.get_iter_first(out iter)){
			selection.select_iter(iter);
		}
		selector.grab_focus();
	}
	
	/**
	 * Called when extension is selected.
	 */
	private void on_selection_changed(Gtk.TreeSelection selection){
		Gtk.TreeIter iter;
		string id;
		bool enabled;
		if(selection.get_selected(null, out iter)){
			model.get(iter, EMI.ID, out id, EMI.ACTIVE, out enabled);
			show_info(id, enabled);
		}
	}
	
	/**
	 * Shows information about extension
	 *
	 * @param id			id of extension
	 * @param enabled		whether the extension is currently enabled
	 */
	private void show_info(string id, bool enabled){
		var info = extensions.available_extensions.lookup(id);
		
		
		var text = Markup.printf_escaped("<page><p><b>%s:</b> %s<br /><b>%s:</b> %s<br/><b>%s:</b> %s</p>",
			/// label for name of extension
			_("Name"), info.name, 
			/// label for version of extension
			_("Version"), info.version,
			/// label for author of extension
			_("Author"), info.author) +  info.description + "</page>";
		var buffer = description.get_buffer() as Diorite.SimpleDocBuffer;
		if (buffer != null)
		{
			try
			{
				buffer.load(text);
			}
			catch(MarkupError e)
			{
				warning("Parse error in description: %s", e.message);
				buffer.set_text(text);
			}
		}
		else
		{
			critical("Wrong type of buffer");
			description.get_buffer().set_text(text);
		}
		settings_button.sensitive = false;
		
		if (enabled)
		{
			extension = extensions[id];
			if (extension != null)
				settings_button.sensitive = extension.has_preferences;
		}
		else
			extension = null;
	}
	
	private void on_settings_button_clicked()
	{
		return_if_fail(extension != null);
		var preferences = extension.get_preferences();
		return_if_fail(preferences != null);
		
		var dialog = new Gtk.Dialog.with_buttons(_("Extension settings"),
			get_toplevel() as Gtk.Window, Gtk.DialogFlags.MODAL,
			Gtk.Stock.CLOSE, 1);
		var box = dialog.get_content_area() as Gtk.Box;
		dialog.border_width = 10;
		box.add(preferences);
		box.show_all();
		dialog.run();
		box.remove(preferences);
		dialog.destroy();
	}
	
	/**
	 * Called when extension is turned on/off.
	 */
	private void on_extension_toggled(string path)
	{
		Gtk.TreeIter iter;
		var treepath =  new Gtk.TreePath.from_string(path);
		string id;
		bool enabled;
		if (model.get_iter(out iter, treepath))
		{
			model.get(iter, EMI.ID, out id, EMI.ACTIVE, out enabled);
			if (enabled)
			{
				if (extensions.unload(id))
					enabled = !enabled;
			}
			else
			{
				if (extensions.load(id) != null)
					enabled = !enabled;
			}
			model.set(iter, EMI.ACTIVE, enabled);
			config.set_bool("extension.%s.enabled".printf(id), enabled);
			show_info(id, enabled);
		}
	}
}


private enum EMI{
	ID,
	ACTIVE,
	NAME,
	COUNT;
}

} // namespace Nuvola
