/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Dialog with information about supported formats.
 */

public class Nuvola.FormatSupportDialog: Gtk.Window
{
	/**
	 * Creates new FormatSupportDialog.
	 * 
	 * @param flash_available    whether Flash plugin is available
	 * @param mp3_available      whether GStreamer MP3 support is available
	 * @param storage            storage to load tests from
	 * @param parent             parent window
	 */
	public FormatSupportDialog(bool flash_available, bool mp3_available, Diorite.Storage storage, Gtk.Window? parent=null)
	{
		/// Title of the dialog that shows information about supported formats for playback
		GLib.Object(title: _("Format support"), window_position: Gtk.WindowPosition.CENTER);
		
		var notebook = new Gtk.Notebook();
		
		var grid = new Gtk.Grid();
		Gtk.Label label;
		var row = 0;
		
		/// "<i>" & "</i>" are formating marks and must be preserved
		label = new Diorite.WrappedLabel(_("Streaming services require <i>Flash</i> or <i>HTML5 Audio</i> for music playback. Service selector screen provides information about which technology is necessary for a particular service."), true);
		label.margin = 15;
		label.set_size_request(300, -1);
		grid.attach(label, 0, row, 2, 1);
		row++;
		
		var icon = new Gtk.Image.from_stock(flash_available ? Gtk.Stock.YES : Gtk.Stock.NO, Gtk.IconSize.BUTTON);
		grid.attach(icon, 0, row, 1, 1);
		/// Label in the Format support dialog
		label = new Gtk.Label("<b>%s</b>".printf(Markup.escape_text(_("Adobe Flash support"))));
		label.use_markup = true;
		label.halign = Gtk.Align.START;
		grid.attach(label, 1, row, 1, 1);
		row++;
		if (flash_available)
			label = new Diorite.WrappedLabel(Markup.escape_text(
			_("Adobe Flash plugin compatible with GTK+ 3 has been successfully detected on your system.")));
		else
			label = new Diorite.WrappedLabel(Markup.escape_text(
			_("Adobe Flash plugin compatible with GTK+ 3 has not been found on your system. Some services may fail to play music."))
			+ "\n\n<a href='%s'>%s</a>".printf(FormatSupport.FLASH_SUPPORT_URL, Markup.escape_text(
			_("How to install compatible Flash plugin."))), true);
		label.margin = 15;
		label.set_size_request(300, -1);
		grid.attach(label, 0, row, 2, 1);
		row++;

		icon = new Gtk.Image.from_stock(mp3_available ? Gtk.Stock.YES : Gtk.Stock.NO, Gtk.IconSize.BUTTON);
		grid.attach(icon, 0, row, 1, 1);
		/// Label in the Format support dialog
		label = new Gtk.Label("<b>%s</b>".printf(Markup.escape_text(_("HTML Audio support"))));
		label.use_markup = true;
		label.halign = Gtk.Align.START;
		grid.attach(label, 1, row, 1, 1);
		row++;
		if (mp3_available)
			label = new Diorite.WrappedLabel(Markup.escape_text(
			_("GStreamer with MP3 support has been successfully detected on your system.")));
		else
			label = new Diorite.WrappedLabel(Markup.escape_text(
			_("GStreamer with MP3 support has not been found on your system. Some services may fail to play music."))
			+ "\n\n<a href='%s'>%s</a>".printf(FormatSupport.MP3_SUPPORT_URL, Markup.escape_text(
			_("How to install GStreamer with MP3 support."))), true);
		label.margin = 15;
		label.set_size_request(300, -1);
		grid.attach(label, 0, row, 2, 1);
		row++;
		var buttons = new Gtk.Grid();
		var close_button = new Gtk.Button.from_stock(Gtk.Stock.CLOSE);
		close_button.clicked.connect(on_close);
		close_button.valign = Gtk.Align.END;
		close_button.margin = 10;
		buttons.add(close_button);
		buttons.hexpand = false;
		buttons.halign = Gtk.Align.END;
		grid.show_all();
		
		/// tab name in format support dialog
		notebook.append_page(grid, new Gtk.Label(_("Status")));
		
		var test = storage.get_data_file("html/Audio.html");
		var view = new StaticWebView(test);
		var scroll = new Gtk.ScrolledWindow(null, null);
		scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
		scroll.add(view);
		scroll.show_all();
		/// tab name in format support dialog
		notebook.append_page(scroll, new Gtk.Label(_("HTML5 Audio")));
		
		test = storage.get_data_file("html/Flash.html");
		view = new StaticWebView(test, true);
		scroll = new Gtk.ScrolledWindow(null, null);
		scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
		scroll.add(view);
		scroll.show_all();
		/// tab name in format support dialog
		notebook.append_page(scroll, new Gtk.Label(_("Flash plugin")));
		
		var main_grid = new Gtk.Grid();
		main_grid.orientation = Gtk.Orientation.VERTICAL;
		main_grid.add(notebook);
		main_grid.add(buttons);
		add(main_grid);
		main_grid.show_all();
		
		
		if (parent != null)
			set_transient_for(parent);
		resizable = false;
	}
	
	private void on_close()
	{
		destroy();
	}
	
	#if DEBUG_MEMORY
	~FormatSupportDialog()
	{
		debug("~FormatSupportDialog");
	}
	#endif
}
